import asyncio
import base64
import copy
import json
import socket
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from urllib import parse

import grpc
from grpc._cython import cygrpc

from pymilvus.client.types import GrantInfo, ResourceGroupConfig
from pymilvus.decorators import ignore_unimplemented, retry_on_rpc_failure
from pymilvus.exceptions import (
    AmbiguousIndexName,
    DataNotMatchException,
    DescribeCollectionException,
    ErrorCode,
    ExceptionsMessage,
    MilvusException,
    ParamError,
)
from pymilvus.grpc_gen import common_pb2, milvus_pb2_grpc
from pymilvus.grpc_gen import milvus_pb2 as milvus_types
from pymilvus.orm.schema import Function
from pymilvus.settings import Config

from . import entity_helper, ts_utils, utils
from .abstract import AnnSearchRequest, BaseRanker, CollectionSchema, FieldSchema, MutationResult
from .async_interceptor import async_header_adder_interceptor
from .check import (
    check_pass_param,
    is_legal_host,
    is_legal_port,
)
from .constants import ITERATOR_SESSION_TS_FIELD
from .interceptor import _api_level_md
from .prepare import Prepare
from .search_result import SearchResult
from .types import (
    AnalyzeResult,
    CompactionState,
    DatabaseInfo,
    DataType,
    HybridExtraList,
    IndexState,
    LoadState,
    ReplicaInfo,
    Shard,
    State,
    Status,
    get_extra_info,
)
from .utils import (
    check_invalid_binary_vector,
    check_status,
    get_server_type,
    is_successful,
    len_of,
)


class AsyncGrpcHandler:
    def __init__(
        self,
        uri: str = Config.GRPC_URI,
        host: str = "",
        port: str = "",
        channel: Optional[grpc.aio.Channel] = None,
        **kwargs,
    ) -> None:
        self._async_stub = None
        self._async_channel = channel
        self.schema_cache: Dict[str, dict] = {}

        addr = kwargs.get("address")
        self._address = addr if addr is not None else self.__get_address(uri, host, port)
        self._log_level = None
        self._user = kwargs.get("user")
        self._set_authorization(**kwargs)
        self._setup_db_name(kwargs.get("db_name"))
        self._setup_grpc_channel(**kwargs)
        self._is_channel_ready = False
        self.callbacks = []  # Do nothing

    def __get_address(self, uri: str, host: str, port: str) -> str:
        if host != "" and port != "" and is_legal_host(host) and is_legal_port(port):
            return f"{host}:{port}"

        try:
            parsed_uri = parse.urlparse(uri)
        except Exception as e:
            raise ParamError(message=f"Illegal uri: [{uri}], {e}") from e
        return parsed_uri.netloc

    def _set_authorization(self, **kwargs):
        secure = kwargs.get("secure", False)
        if not isinstance(secure, bool):
            raise ParamError(message="secure must be bool type")
        self._secure = secure
        self._client_pem_path = kwargs.get("client_pem_path", "")
        self._client_key_path = kwargs.get("client_key_path", "")
        self._ca_pem_path = kwargs.get("ca_pem_path", "")
        self._server_pem_path = kwargs.get("server_pem_path", "")
        self._server_name = kwargs.get("server_name", "")

        self._async_authorization_interceptor = None

    def __enter__(self):
        return self

    def __exit__(self: object, exc_type: object, exc_val: object, exc_tb: object):
        pass

    async def close(self):
        await self._async_channel.close()
        self._async_channel = None

    def _setup_authorization_interceptor(self, user: str, password: str, token: str):
        keys = []
        values = []
        if token:
            authorization = base64.b64encode(f"{token}".encode())
            keys.append("authorization")
            values.append(authorization)
        elif user and password:
            authorization = base64.b64encode(f"{user}:{password}".encode())
            keys.append("authorization")
            values.append(authorization)
        if len(keys) > 0 and len(values) > 0:
            self._async_authorization_interceptor = async_header_adder_interceptor(keys, values)
            self._final_channel._unary_unary_interceptors.append(
                self._async_authorization_interceptor
            )

    def _setup_db_name(self, db_name: str):
        if db_name is None:
            new_db = None
        else:
            check_pass_param(db_name=db_name)
            new_db = db_name

        if getattr(self, "_db_name", None) != new_db:
            self.schema_cache.clear()

        self._db_name = new_db

    def _setup_grpc_channel(self, **kwargs):
        if self._async_channel is None:
            opts = [
                (cygrpc.ChannelArgKey.max_send_message_length, -1),
                (cygrpc.ChannelArgKey.max_receive_message_length, -1),
                ("grpc.enable_retries", 1),
                ("grpc.keepalive_time_ms", 55000),
            ]
            if not self._secure:
                self._async_channel = grpc.aio.insecure_channel(
                    self._address,
                    options=opts,
                )
            else:
                if self._server_name != "":
                    opts.append(("grpc.ssl_target_name_override", self._server_name))

                root_cert, private_k, cert_chain = None, None, None
                if self._server_pem_path != "":
                    with Path(self._server_pem_path).open("rb") as f:
                        root_cert = f.read()
                elif (
                    self._client_pem_path != ""
                    and self._client_key_path != ""
                    and self._ca_pem_path != ""
                ):
                    with Path(self._ca_pem_path).open("rb") as f:
                        root_cert = f.read()
                    with Path(self._client_key_path).open("rb") as f:
                        private_k = f.read()
                    with Path(self._client_pem_path).open("rb") as f:
                        cert_chain = f.read()

                creds = grpc.ssl_channel_credentials(
                    root_certificates=root_cert,
                    private_key=private_k,
                    certificate_chain=cert_chain,
                )
                self._async_channel = grpc.aio.secure_channel(
                    self._address,
                    creds,
                    options=opts,
                )

        # avoid to add duplicate headers.
        self._final_channel = self._async_channel

        if self._async_authorization_interceptor:
            self._final_channel._unary_unary_interceptors.append(
                self._async_authorization_interceptor
            )
        else:
            self._setup_authorization_interceptor(
                kwargs.get("user"),
                kwargs.get("password"),
                kwargs.get("token"),
            )
        if self._db_name:
            async_db_interceptor = async_header_adder_interceptor(["dbname"], [self._db_name])
            self._final_channel._unary_unary_interceptors.append(async_db_interceptor)
        if self._log_level:
            async_log_level_interceptor = async_header_adder_interceptor(
                ["log-level"], [self._log_level]
            )
            self._final_channel._unary_unary_interceptors.append(async_log_level_interceptor)
            self._log_level = None
        self._async_stub = milvus_pb2_grpc.MilvusServiceStub(self._final_channel)

    @property
    def server_address(self):
        return self._address

    def get_server_type(self):
        return get_server_type(self.server_address.split(":")[0])

    async def ensure_channel_ready(self):
        try:
            if not self._is_channel_ready:
                # wait for channel ready
                await self._async_channel.channel_ready()
                # set identifier interceptor
                host = socket.gethostname()
                req = Prepare.register_request(self._user, host)
                response = await self._async_stub.Connect(request=req)
                check_status(response.status)
                _async_identifier_interceptor = async_header_adder_interceptor(
                    ["identifier"], [str(response.identifier)]
                )
                self._async_channel._unary_unary_interceptors.append(_async_identifier_interceptor)
                self._async_stub = milvus_pb2_grpc.MilvusServiceStub(self._async_channel)

                self._is_channel_ready = True
        except grpc.FutureTimeoutError as e:
            raise MilvusException(
                code=Status.CONNECT_FAILED,
                message=f"Fail connecting to server on {self._address}, illegal connection params or server unavailable",
            ) from e
        except Exception as e:
            raise e from e

    @retry_on_rpc_failure()
    async def create_collection(
        self, collection_name: str, fields: List, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.create_collection_request(collection_name, fields, **kwargs)
        response = await self._async_stub.CreateCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def drop_collection(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.drop_collection_request(collection_name)
        response = await self._async_stub.DropCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def load_collection(
        self,
        collection_name: str,
        replica_number: Optional[int] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()

        check_pass_param(timeout=timeout)
        request = Prepare.load_collection(collection_name, replica_number, **kwargs)
        response = await self._async_stub.LoadCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

        await self.wait_for_loading_collection(
            collection_name=collection_name,
            is_refresh=request.refresh,
            timeout=timeout,
            **kwargs,
        )

    @retry_on_rpc_failure()
    async def wait_for_loading_collection(
        self,
        collection_name: str,
        timeout: Optional[float] = None,
        is_refresh: bool = False,
        **kwargs,
    ):
        start = time.time()

        def can_loop(t: int) -> bool:
            return True if timeout is None else t <= (start + timeout)

        while can_loop(time.time()):
            progress = await self.get_loading_progress(
                collection_name=collection_name,
                is_refresh=is_refresh,
                timeout=timeout,
                **kwargs,
            )
            if progress >= 100:
                return
            await asyncio.sleep(Config.WaitTimeDurationWhenLoad)
        raise MilvusException(
            message=f"wait for loading collection timeout, collection: {collection_name}"
        )

    @retry_on_rpc_failure()
    async def get_loading_progress(
        self,
        collection_name: str,
        partition_names: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        is_refresh: bool = False,
        **kwargs,
    ):
        request = Prepare.get_loading_progress(collection_name, partition_names)
        response = await self._async_stub.GetLoadingProgress(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        if is_refresh:
            return response.refresh_progress
        return response.progress

    @retry_on_rpc_failure()
    async def describe_collection(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.describe_collection_request(collection_name)
        response = await self._async_stub.DescribeCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status

        if is_successful(status):
            return CollectionSchema(raw=response).dict()

        raise DescribeCollectionException(status.code, status.reason, status.error_code)

    @retry_on_rpc_failure()
    async def has_collection(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ) -> bool:
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.describe_collection_request(collection_name)
        reply = await self._async_stub.DescribeCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )

        if (
            reply.status.error_code == common_pb2.UnexpectedError
            and "can't find collection" in reply.status.reason
        ):
            return False

        if reply.status.error_code == common_pb2.CollectionNotExists:
            return False

        if is_successful(reply.status):
            return True

        if reply.status.code == ErrorCode.COLLECTION_NOT_FOUND:
            return False

        raise MilvusException(reply.status.code, reply.status.reason, reply.status.error_code)

    @retry_on_rpc_failure()
    async def list_collections(self, timeout: Optional[float] = None, **kwargs) -> List[str]:
        await self.ensure_channel_ready()
        request = Prepare.show_collections_request()
        response = await self._async_stub.ShowCollections(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        check_status(status)
        return list(response.collection_names)

    @retry_on_rpc_failure()
    async def get_collection_stats(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        index_param = Prepare.get_collection_stats_request(collection_name)
        response = await self._async_stub.GetCollectionStatistics(
            index_param, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        check_status(status)
        return response.stats

    @retry_on_rpc_failure()
    async def get_partition_stats(
        self, collection_name: str, partition_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        req = Prepare.get_partition_stats_request(collection_name, partition_name)
        response = await self._async_stub.GetPartitionStatistics(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        check_status(status)
        return response.stats

    @retry_on_rpc_failure()
    async def get_load_state(
        self,
        collection_name: str,
        partition_names: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        request = Prepare.get_load_state(collection_name, partition_names)
        response = await self._async_stub.GetLoadState(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return LoadState(response.state)

    @retry_on_rpc_failure()
    async def refresh_load(
        self,
        collection_name: str,
        partition_names: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        request = Prepare.get_loading_progress(collection_name, partition_names)
        response = await self._async_stub.GetLoadingProgress(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return response.refresh_progress

    @retry_on_rpc_failure()
    async def get_server_version(self, timeout: Optional[float] = None, **kwargs) -> str:
        await self.ensure_channel_ready()
        req = Prepare.get_server_version()
        resp = await self._async_stub.GetVersion(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.version

    @retry_on_rpc_failure()
    async def describe_replica(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ) -> List[ReplicaInfo]:
        await self.ensure_channel_ready()
        collection_id = (await self.describe_collection(collection_name, timeout, **kwargs))[
            "collection_id"
        ]

        req = Prepare.get_replicas(collection_id)
        response = await self._async_stub.GetReplicas(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)

        groups = []
        for replica in response.replicas:
            shards = [
                Shard(s.dm_channel_name, s.node_ids, s.leaderID) for s in replica.shard_replicas
            ]
            groups.append(
                ReplicaInfo(
                    replica.replicaID,
                    shards,
                    replica.node_ids,
                    replica.resource_group_name,
                    replica.num_outbound_node,
                )
            )

        return groups

    @retry_on_rpc_failure()
    async def rename_collection(
        self,
        old_name: str,
        new_name: str,
        new_db_name: str = "",
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=new_name, timeout=timeout)
        check_pass_param(collection_name=old_name)
        if new_db_name:
            check_pass_param(db_name=new_db_name)
        request = Prepare.rename_collections_request(old_name, new_name, new_db_name)
        status = await self._async_stub.RenameCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    async def _get_info(self, collection_name: str, timeout: Optional[float] = None, **kwargs):
        schema = kwargs.get("schema")
        schema, schema_timestamp = await self._get_schema_from_cache_or_remote(
            collection_name, schema=schema, timeout=timeout, **kwargs
        )
        fields_info = schema.get("fields")
        struct_fields_info = schema.get("struct_array_fields", [])
        enable_dynamic = schema.get("enable_dynamic_field", False)

        return fields_info, struct_fields_info, enable_dynamic, schema_timestamp

    async def update_schema(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ) -> dict:
        self.schema_cache.pop(collection_name, None)
        schema = await self.describe_collection(collection_name, timeout=timeout, **kwargs)
        schema_timestamp = schema.get("update_timestamp", 0)
        self.schema_cache[collection_name] = {
            "schema": schema,
            "schema_timestamp": schema_timestamp,
        }
        return schema

    async def _get_schema_from_cache_or_remote(
        self,
        collection_name: str,
        schema: Optional[dict] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Tuple[dict, int]:
        if collection_name in self.schema_cache:
            cached = self.schema_cache[collection_name]
            return cached["schema"], cached["schema_timestamp"]

        if not isinstance(schema, dict):
            schema = await self.describe_collection(collection_name, timeout=timeout, **kwargs)
        schema_timestamp = schema.get("update_timestamp", 0)
        self.schema_cache[collection_name] = {
            "schema": schema,
            "schema_timestamp": schema_timestamp,
        }
        return schema, schema_timestamp

    @retry_on_rpc_failure()
    async def release_collection(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.release_collection("", collection_name)
        response = await self._async_stub.ReleaseCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def insert_rows(
        self,
        collection_name: str,
        entities: Union[Dict, List[Dict]],
        partition_name: Optional[str] = None,
        schema: Optional[dict] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        try:
            request = await self._prepare_row_insert_request(
                collection_name, entities, partition_name, schema, timeout, **kwargs
            )
        except DataNotMatchException:
            schema = await self.update_schema(collection_name, timeout, **kwargs)
            request = await self._prepare_row_insert_request(
                collection_name, entities, partition_name, schema, timeout, **kwargs
            )
        resp = await self._async_stub.Insert(
            request=request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        if resp.status.error_code == common_pb2.SchemaMismatch:
            schema = await self.update_schema(collection_name, timeout, **kwargs)
            request = await self._prepare_row_insert_request(
                collection_name, entities, partition_name, schema, timeout, **kwargs
            )
            resp = await self._async_stub.Insert(
                request=request, timeout=timeout, metadata=_api_level_md(**kwargs)
            )
        check_status(resp.status)
        ts_utils.update_collection_ts(collection_name, resp.timestamp)
        return MutationResult(resp)

    async def _prepare_row_insert_request(
        self,
        collection_name: str,
        entity_rows: Union[List[Dict], Dict],
        partition_name: Optional[str] = None,
        schema: Optional[dict] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        if isinstance(entity_rows, dict):
            entity_rows = [entity_rows]

        schema, schema_timestamp = await self._get_schema_from_cache_or_remote(
            collection_name, schema=schema, timeout=timeout, **kwargs
        )
        fields_info = schema.get("fields")
        struct_fields_info = schema.get("struct_array_fields", [])  # Default to empty list
        enable_dynamic = schema.get("enable_dynamic_field", False)

        return Prepare.row_insert_param(
            collection_name,
            entity_rows,
            partition_name,
            fields_info,
            struct_fields_info,
            enable_dynamic=enable_dynamic,
            schema_timestamp=schema_timestamp,
        )

    @retry_on_rpc_failure()
    async def delete(
        self,
        collection_name: str,
        expression: str,
        partition_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        try:
            req = Prepare.delete_request(
                collection_name=collection_name,
                filter=expression,
                partition_name=partition_name,
                consistency_level=kwargs.pop("consistency_level", 0),
                **kwargs,
            )

            response = await self._async_stub.Delete(
                req, timeout=timeout, metadata=_api_level_md(**kwargs)
            )

            m = MutationResult(response)
            ts_utils.update_collection_ts(collection_name, m.timestamp)
        except Exception as err:
            raise err from err
        else:
            return m

    async def _prepare_batch_upsert_request(
        self,
        collection_name: str,
        entities: List,
        partition_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        param = kwargs.get("upsert_param")
        if param and not isinstance(param, milvus_types.UpsertRequest):
            raise ParamError(message="The value of key 'upsert_param' is invalid")
        if not isinstance(entities, list):
            raise ParamError(message="'entities' must be a list, please provide valid entity data.")

        # Extract partial_update parameter from kwargs
        partial_update = kwargs.get("partial_update", False)

        schema = kwargs.get("schema")
        schema, _ = await self._get_schema_from_cache_or_remote(
            collection_name, schema=schema, timeout=timeout, **kwargs
        )

        fields_info = schema["fields"]

        return (
            param
            if param
            else Prepare.batch_upsert_param(
                collection_name,
                entities,
                partition_name,
                fields_info,
                partial_update=partial_update,
            )
        )

    @retry_on_rpc_failure()
    async def upsert(
        self,
        collection_name: str,
        entities: List,
        partition_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        if not check_invalid_binary_vector(entities):
            raise ParamError(message="Invalid binary vector data exists")

        request = await self._prepare_batch_upsert_request(
            collection_name, entities, partition_name, timeout, **kwargs
        )
        response = await self._async_stub.Upsert(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        m = MutationResult(response)
        ts_utils.update_collection_ts(collection_name, m.timestamp)
        return m

    async def _prepare_row_upsert_request(
        self,
        collection_name: str,
        rows: List,
        partition_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        if not isinstance(rows, list):
            raise ParamError(message="'rows' must be a list, please provide valid row data.")

        partial_update = kwargs.get("partial_update", False)
        fields_info, struct_fields_info, enable_dynamic, schema_timestamp = await self._get_info(
            collection_name, timeout, **kwargs
        )
        return Prepare.row_upsert_param(
            collection_name,
            rows,
            partition_name,
            fields_info,
            struct_fields_info,
            enable_dynamic=enable_dynamic,
            partial_update=partial_update,
            schema_timestamp=schema_timestamp,
        )

    @retry_on_rpc_failure()
    async def upsert_rows(
        self,
        collection_name: str,
        entities: List,
        partition_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        if isinstance(entities, dict):
            entities = [entities]
        try:
            request = await self._prepare_row_upsert_request(
                collection_name, entities, partition_name, timeout, **kwargs
            )
        except DataNotMatchException:
            schema = await self.update_schema(collection_name, timeout, **kwargs)
            request = await self._prepare_row_upsert_request(
                collection_name, entities, partition_name, timeout, schema=schema, **kwargs
            )
        response = await self._async_stub.Upsert(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        if response.status.error_code == common_pb2.SchemaMismatch:
            schema = await self.update_schema(collection_name, timeout, **kwargs)
            request = await self._prepare_row_upsert_request(
                collection_name, entities, partition_name, timeout, schema=schema, **kwargs
            )
            response = await self._async_stub.Upsert(
                request, timeout=timeout, metadata=_api_level_md(**kwargs)
            )
        check_status(response.status)
        m = MutationResult(response)
        ts_utils.update_collection_ts(collection_name, m.timestamp)
        return m

    async def _execute_search(
        self, request: milvus_types.SearchRequest, timeout: Optional[float] = None, **kwargs
    ):
        response = await self._async_stub.Search(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        round_decimal = kwargs.get("round_decimal", -1)
        return SearchResult(
            response.results,
            round_decimal,
            status=response.status,
            session_ts=response.session_ts,
        )

    async def _execute_hybrid_search(
        self, request: milvus_types.HybridSearchRequest, timeout: Optional[float] = None, **kwargs
    ):
        response = await self._async_stub.HybridSearch(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        round_decimal = kwargs.get("round_decimal", -1)
        return SearchResult(response.results, round_decimal, status=response.status)

    @retry_on_rpc_failure()
    async def search(
        self,
        collection_name: str,
        data: Union[List[List[float]], utils.SparseMatrixInputType],
        anns_field: str,
        param: Dict,
        limit: int,
        expression: Optional[str] = None,
        partition_names: Optional[List[str]] = None,
        output_fields: Optional[List[str]] = None,
        round_decimal: int = -1,
        timeout: Optional[float] = None,
        ranker: Optional[Function] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(
            limit=limit,
            round_decimal=round_decimal,
            anns_field=anns_field,
            search_data=data,
            partition_name_array=partition_names,
            output_fields=output_fields,
            guarantee_timestamp=kwargs.get("guarantee_timestamp"),
            timeout=timeout,
        )
        request = Prepare.search_requests_with_expr(
            collection_name,
            data,
            anns_field,
            param,
            limit,
            expression,
            partition_names,
            output_fields,
            round_decimal,
            ranker=ranker,
            **kwargs,
        )
        return await self._execute_search(request, timeout, round_decimal=round_decimal, **kwargs)

    @retry_on_rpc_failure()
    async def hybrid_search(
        self,
        collection_name: str,
        reqs: List[AnnSearchRequest],
        rerank: Union[BaseRanker, Function],
        limit: int,
        partition_names: Optional[List[str]] = None,
        output_fields: Optional[List[str]] = None,
        round_decimal: int = -1,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(
            limit=limit,
            round_decimal=round_decimal,
            partition_name_array=partition_names,
            output_fields=output_fields,
            guarantee_timestamp=kwargs.get("guarantee_timestamp"),
            timeout=timeout,
        )

        requests = []
        for req in reqs:
            search_request = Prepare.search_requests_with_expr(
                collection_name,
                req.data,
                req.anns_field,
                req.param,
                req.limit,
                req.expr,
                partition_names=partition_names,
                round_decimal=round_decimal,
                expr_params=req.expr_params,
                **kwargs,
            )
            requests.append(search_request)

        hybrid_search_request = Prepare.hybrid_search_request_with_ranker(
            collection_name,
            requests,
            rerank,
            limit,
            partition_names,
            output_fields,
            round_decimal,
            **kwargs,
        )
        return await self._execute_hybrid_search(
            hybrid_search_request, timeout, round_decimal=round_decimal, **kwargs
        )

    @retry_on_rpc_failure()
    async def create_index(
        self,
        collection_name: str,
        field_name: str,
        params: Dict,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        index_name = kwargs.pop("index_name", Config.IndexName)
        copy_kwargs = copy.deepcopy(kwargs)

        collection_desc = await self.describe_collection(
            collection_name, timeout=timeout, **copy_kwargs
        )
        await self.ensure_channel_ready()
        valid_field = False
        for fields in collection_desc["fields"]:
            if field_name != fields["name"]:
                continue
            valid_field = True
            if fields["type"] not in {
                DataType.FLOAT_VECTOR,
                DataType.BINARY_VECTOR,
                DataType.FLOAT16_VECTOR,
                DataType.BFLOAT16_VECTOR,
                DataType.SPARSE_FLOAT_VECTOR,
                DataType.INT8_VECTOR,
            }:
                break

        if not valid_field:
            raise MilvusException(message=f"cannot create index on non-existed field: {field_name}")

        index_param = Prepare.create_index_request(
            collection_name, field_name, params, index_name=index_name
        )

        status = await self._async_stub.CreateIndex(
            index_param, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

        index_success, fail_reason = await self.wait_for_creating_index(
            collection_name=collection_name,
            index_name=index_name,
            timeout=timeout,
            field_name=field_name,
            **kwargs,
        )

        if not index_success:
            raise MilvusException(message=fail_reason)

        return Status(status.code, status.reason)

    @retry_on_rpc_failure()
    async def wait_for_creating_index(
        self, collection_name: str, index_name: str, timeout: Optional[float] = None, **kwargs
    ):
        timestamp = await self.alloc_timestamp()
        start = time.time()
        while True:
            await asyncio.sleep(0.5)
            state, fail_reason = await self.get_index_state(
                collection_name, index_name, timeout=timeout, timestamp=timestamp, **kwargs
            )
            if state == IndexState.Finished:
                return True, fail_reason
            if state == IndexState.Failed:
                return False, fail_reason
            end = time.time()
            if isinstance(timeout, int) and end - start > timeout:
                msg = (
                    f"collection {collection_name} create index {index_name} "
                    f"timeout in {timeout}s"
                )
                raise MilvusException(message=msg)

    @retry_on_rpc_failure()
    async def get_index_state(
        self,
        collection_name: str,
        index_name: str,
        timeout: Optional[float] = None,
        timestamp: Optional[int] = None,
        **kwargs,
    ):
        request = Prepare.describe_index_request(collection_name, index_name, timestamp)
        response = await self._async_stub.DescribeIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        check_status(status)

        if len(response.index_descriptions) == 1:
            index_desc = response.index_descriptions[0]
            return index_desc.state, index_desc.index_state_fail_reason
        field_name = kwargs.pop("field_name", "")
        if field_name != "":
            for index_desc in response.index_descriptions:
                if index_desc.field_name == field_name:
                    return index_desc.state, index_desc.index_state_fail_reason

        raise AmbiguousIndexName(message=ExceptionsMessage.AmbiguousIndexName)

    @retry_on_rpc_failure()
    async def drop_index(
        self,
        collection_name: str,
        field_name: str,
        index_name: str,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.drop_index_request(collection_name, field_name, index_name)
        response = await self._async_stub.DropIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def create_partition(
        self, collection_name: str, partition_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(
            collection_name=collection_name, partition_name=partition_name, timeout=timeout
        )
        request = Prepare.create_partition_request(collection_name, partition_name)
        response = await self._async_stub.CreatePartition(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def drop_partition(
        self, collection_name: str, partition_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(
            collection_name=collection_name, partition_name=partition_name, timeout=timeout
        )
        request = Prepare.drop_partition_request(collection_name, partition_name)

        response = await self._async_stub.DropPartition(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def load_partitions(
        self,
        collection_name: str,
        partition_names: List[str],
        replica_number: Optional[int] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(timeout=timeout)

        request = Prepare.load_partitions(
            collection_name=collection_name,
            partition_names=partition_names,
            replica_number=replica_number,
            **kwargs,
        )
        response = await self._async_stub.LoadPartitions(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

        await self.wait_for_loading_partitions(
            collection_name=collection_name,
            partition_names=partition_names,
            is_refresh=request.refresh,
            timeout=timeout,
            **kwargs,
        )

    @retry_on_rpc_failure()
    async def wait_for_loading_partitions(
        self,
        collection_name: str,
        partition_names: List[str],
        timeout: Optional[float] = None,
        is_refresh: bool = False,
        **kwargs,
    ):
        start = time.time()

        def can_loop(t: int) -> bool:
            return True if timeout is None else t <= (start + timeout)

        while can_loop(time.time()):
            progress = await self.get_loading_progress(
                collection_name, partition_names, timeout=timeout, is_refresh=is_refresh, **kwargs
            )
            if progress >= 100:
                return
            await asyncio.sleep(Config.WaitTimeDurationWhenLoad)
        raise MilvusException(
            message=f"wait for loading partition timeout, collection: {collection_name}, partitions: {partition_names}"
        )

    @retry_on_rpc_failure()
    async def release_partitions(
        self,
        collection_name: str,
        partition_names: List[str],
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(
            collection_name=collection_name, partition_name_array=partition_names, timeout=timeout
        )
        request = Prepare.release_partitions("", collection_name, partition_names)
        response = await self._async_stub.ReleasePartitions(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def has_partition(
        self, collection_name: str, partition_name: str, timeout: Optional[float] = None, **kwargs
    ) -> bool:
        await self.ensure_channel_ready()
        check_pass_param(
            collection_name=collection_name, partition_name=partition_name, timeout=timeout
        )
        request = Prepare.has_partition_request(collection_name, partition_name)
        response = await self._async_stub.HasPartition(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return response.value

    @retry_on_rpc_failure()
    async def list_partitions(
        self, collection_name: str, timeout: Optional[float] = None, **kwargs
    ) -> List[str]:
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.show_partitions_request(collection_name)
        response = await self._async_stub.ShowPartitions(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return list(response.partition_names)

    @retry_on_rpc_failure()
    async def get(
        self,
        collection_name: str,
        ids: List[int],
        output_fields: Optional[List[str]] = None,
        partition_names: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        # TODO: some check
        await self.ensure_channel_ready()
        request = Prepare.retrieve_request(collection_name, ids, output_fields, partition_names)
        return await self._async_stub.Retrieve.get(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )

    @retry_on_rpc_failure()
    async def query(
        self,
        collection_name: str,
        expr: str,
        output_fields: Optional[List[str]] = None,
        partition_names: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        strict_float32: bool = False,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        if output_fields is not None and not isinstance(output_fields, (list,)):
            raise ParamError(message="Invalid query format. 'output_fields' must be a list")
        request = Prepare.query_request(
            collection_name, expr, output_fields, partition_names, **kwargs
        )
        response = await self._async_stub.Query(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)

        num_fields = len(response.fields_data)
        # check has fields
        if num_fields == 0:
            raise MilvusException(message="No fields returned")

        # check if all lists are of the same length
        it = iter(response.fields_data)
        num_entities = len_of(next(it))
        if not all(len_of(field_data) == num_entities for field_data in it):
            raise MilvusException(message="The length of fields data is inconsistent")

        _, dynamic_fields = entity_helper.extract_dynamic_field_from_result(response)
        keys = [field_data.field_name for field_data in response.fields_data]
        filtered_keys = [k for k in keys if k != "$meta"]
        results = [dict.fromkeys(filtered_keys) for _ in range(num_entities)]
        lazy_field_data = []
        for field_data in response.fields_data:
            lazy_extracted = entity_helper.extract_row_data_from_fields_data_v2(field_data, results)
            if lazy_extracted:
                lazy_field_data.append(field_data)

        extra_dict = get_extra_info(response.status)
        extra_dict[ITERATOR_SESSION_TS_FIELD] = response.session_ts
        return HybridExtraList(
            lazy_field_data,
            results,
            extra=extra_dict,
            dynamic_fields=dynamic_fields,
            strict_float32=strict_float32,
        )

    @retry_on_rpc_failure()
    @ignore_unimplemented(0)
    async def alloc_timestamp(self, timeout: Optional[float] = None, **kwargs) -> int:
        await self.ensure_channel_ready()
        request = milvus_types.AllocTimestampRequest()
        response = await self._async_stub.AllocTimestamp(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return response.timestamp

    @retry_on_rpc_failure()
    async def alter_collection_properties(
        self, collection_name: str, properties: dict, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, properties=properties, timeout=timeout)
        request = Prepare.alter_collection_request(collection_name, properties=properties)
        status = await self._async_stub.AlterCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def drop_collection_properties(
        self,
        collection_name: str,
        property_keys: List[str],
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.alter_collection_request(collection_name, delete_keys=property_keys)
        status = await self._async_stub.AlterCollection(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def alter_collection_field(
        self,
        collection_name: str,
        field_name: str,
        field_params: dict,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, properties=field_params, timeout=timeout)
        request = Prepare.alter_collection_field_request(
            collection_name=collection_name, field_name=field_name, field_param=field_params
        )
        status = await self._async_stub.AlterCollectionField(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def add_collection_field(
        self,
        collection_name: str,
        field_schema: FieldSchema,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.add_collection_field_request(collection_name, field_schema)
        status = await self._async_stub.AddCollectionField(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def list_indexes(self, collection_name: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.describe_index_request(collection_name, "")

        response = await self._async_stub.DescribeIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        if is_successful(status):
            return response.index_descriptions
        if status.code == ErrorCode.INDEX_NOT_FOUND or status.error_code == Status.INDEX_NOT_EXIST:
            return []
        raise MilvusException(status.code, status.reason, status.error_code)

    @retry_on_rpc_failure()
    async def describe_index(
        self,
        collection_name: str,
        index_name: str,
        timeout: Optional[float] = None,
        timestamp: Optional[int] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, index_name=index_name, timeout=timeout)
        request = Prepare.describe_index_request(collection_name, index_name, timestamp=timestamp)

        response = await self._async_stub.DescribeIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        status = response.status
        if status.code == ErrorCode.INDEX_NOT_FOUND or status.error_code == Status.INDEX_NOT_EXIST:
            return None
        check_status(status)
        if len(response.index_descriptions) == 1:
            info_dict = {kv.key: kv.value for kv in response.index_descriptions[0].params}
            info_dict["field_name"] = response.index_descriptions[0].field_name
            info_dict["index_name"] = response.index_descriptions[0].index_name
            if info_dict.get("params"):
                info_dict["params"] = json.loads(info_dict["params"])
            info_dict["total_rows"] = response.index_descriptions[0].total_rows
            info_dict["indexed_rows"] = response.index_descriptions[0].indexed_rows
            info_dict["pending_index_rows"] = response.index_descriptions[0].pending_index_rows
            info_dict["state"] = common_pb2.IndexState.Name(response.index_descriptions[0].state)
            return info_dict

        raise AmbiguousIndexName(message=ExceptionsMessage.AmbiguousIndexName)

    @retry_on_rpc_failure()
    async def alter_index_properties(
        self,
        collection_name: str,
        index_name: str,
        properties: dict,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, index_name=index_name, timeout=timeout)
        if properties is None:
            raise ParamError(message="properties should not be None")

        request = Prepare.alter_index_properties_request(collection_name, index_name, properties)
        response = await self._async_stub.AlterIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def drop_index_properties(
        self,
        collection_name: str,
        index_name: str,
        property_keys: List[str],
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, index_name=index_name, timeout=timeout)
        request = Prepare.drop_index_properties_request(
            collection_name, index_name, delete_keys=property_keys
        )
        response = await self._async_stub.AlterIndex(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def create_alias(
        self, collection_name: str, alias: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.create_alias_request(collection_name, alias)
        response = await self._async_stub.CreateAlias(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def drop_alias(self, alias: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        request = Prepare.drop_alias_request(alias)
        response = await self._async_stub.DropAlias(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def alter_alias(
        self, collection_name: str, alias: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.alter_alias_request(collection_name, alias)
        response = await self._async_stub.AlterAlias(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response)

    @retry_on_rpc_failure()
    async def describe_alias(self, alias: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        check_pass_param(alias=alias, timeout=timeout)
        request = Prepare.describe_alias_request(alias)
        response = await self._async_stub.DescribeAlias(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        ret = {
            "alias": alias,
        }
        if response.collection:
            ret["collection_name"] = response.collection
        if response.db_name:
            ret["db_name"] = response.db_name
        return ret

    @retry_on_rpc_failure()
    async def list_aliases(self, collection_name: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        check_pass_param(collection_name=collection_name, timeout=timeout)
        request = Prepare.list_aliases_request(collection_name)
        response = await self._async_stub.ListAliases(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return response.aliases

    def reset_db_name(self, db_name: str):
        self._setup_db_name(db_name)
        self._setup_grpc_channel()

    @retry_on_rpc_failure()
    async def create_database(
        self,
        db_name: str,
        properties: Optional[dict] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        check_pass_param(db_name=db_name, timeout=timeout)
        request = Prepare.create_database_req(db_name, properties=properties)
        status = await self._async_stub.CreateDatabase(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def drop_database(self, db_name: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        request = Prepare.drop_database_req(db_name)
        status = await self._async_stub.DropDatabase(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def list_database(self, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        check_pass_param(timeout=timeout)
        request = Prepare.list_database_req()
        response = await self._async_stub.ListDatabases(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return list(response.db_names)

    @retry_on_rpc_failure()
    async def alter_database(
        self, db_name: str, properties: dict, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        request = Prepare.alter_database_properties_req(db_name, properties)
        status = await self._async_stub.AlterDatabase(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def drop_database_properties(
        self, db_name: str, property_keys: List[str], timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        request = Prepare.drop_database_properties_req(db_name, property_keys)
        status = await self._async_stub.AlterDatabase(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(status)

    @retry_on_rpc_failure()
    async def describe_database(self, db_name: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        request = Prepare.describe_database_req(db_name=db_name)
        resp = await self._async_stub.DescribeDatabase(
            request, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return DatabaseInfo(resp).to_dict()

    @retry_on_rpc_failure()
    async def create_privilege_group(
        self, privilege_group: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.create_privilege_group_req(privilege_group)
        resp = await self._async_stub.CreatePrivilegeGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def drop_privilege_group(
        self, privilege_group: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.drop_privilege_group_req(privilege_group)
        resp = await self._async_stub.DropPrivilegeGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def list_privilege_groups(self, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        req = Prepare.list_privilege_groups_req()
        resp = await self._async_stub.ListPrivilegeGroups(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.privilege_groups

    @retry_on_rpc_failure()
    async def add_privileges_to_group(
        self, privilege_group: str, privileges: List[str], timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_group_req(
            privilege_group, privileges, milvus_types.OperatePrivilegeGroupType.AddPrivilegesToGroup
        )
        resp = await self._async_stub.OperatePrivilegeGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def remove_privileges_from_group(
        self, privilege_group: str, privileges: List[str], timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_group_req(
            privilege_group,
            privileges,
            milvus_types.OperatePrivilegeGroupType.RemovePrivilegesFromGroup,
        )
        resp = await self._async_stub.OperatePrivilegeGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def create_user(
        self, user: str, password: str, timeout: Optional[float] = None, **kwargs
    ):
        check_pass_param(user=user, password=password, timeout=timeout)
        req = Prepare.create_user_request(user, password)
        resp = await self._async_stub.CreateCredential(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def drop_user(self, user: str, timeout: Optional[float] = None, **kwargs):
        req = Prepare.delete_user_request(user)
        resp = await self._async_stub.DeleteCredential(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def update_password(
        self,
        user: str,
        old_password: str,
        new_password: str,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        req = Prepare.update_password_request(user, old_password, new_password)
        resp = await self._async_stub.UpdateCredential(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def list_users(self, timeout: Optional[float] = None, **kwargs):
        req = Prepare.list_usernames_request()
        resp = await self._async_stub.ListCredUsers(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.usernames

    @retry_on_rpc_failure()
    async def describe_user(
        self, username: str, include_role_info: bool, timeout: Optional[float] = None, **kwargs
    ):
        req = Prepare.select_user_request(username, include_role_info)
        resp = await self._async_stub.SelectUser(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp

    @retry_on_rpc_failure()
    async def create_role(self, role_name: str, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        req = Prepare.create_role_request(role_name)
        resp = await self._async_stub.CreateRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def drop_role(
        self, role_name: str, force_drop: bool = False, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.drop_role_request(role_name, force_drop=force_drop)
        resp = await self._async_stub.DropRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def grant_role(
        self, username: str, role_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_user_role_request(
            username, role_name, milvus_types.OperateUserRoleType.AddUserToRole
        )
        resp = await self._async_stub.OperateUserRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def revoke_role(
        self, username: str, role_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_user_role_request(
            username, role_name, milvus_types.OperateUserRoleType.RemoveUserFromRole
        )
        resp = await self._async_stub.OperateUserRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def describe_role(
        self, role_name: str, include_user_info: bool, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.select_role_request(role_name, include_user_info)
        resp = await self._async_stub.SelectRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.results

    @retry_on_rpc_failure()
    async def list_roles(self, include_user_info: bool, timeout: Optional[float] = None, **kwargs):
        await self.ensure_channel_ready()
        req = Prepare.select_role_request(None, include_user_info)
        resp = await self._async_stub.SelectRole(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.results

    @retry_on_rpc_failure()
    async def select_grant_for_one_role(
        self, role_name: str, db_name: str, timeout: Optional[float] = None, **kwargs
    ):
        await self.ensure_channel_ready()
        req = Prepare.select_grant_request(role_name, None, None, db_name)
        resp = await self._async_stub.SelectGrant(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)

        return GrantInfo(resp.entities)

    @retry_on_rpc_failure()
    async def grant_privilege(
        self,
        role_name: str,
        object: str,
        object_name: str,
        privilege: str,
        db_name: str,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_request(
            role_name,
            object,
            object_name,
            privilege,
            db_name,
            milvus_types.OperatePrivilegeType.Grant,
        )
        resp = await self._async_stub.OperatePrivilege(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def revoke_privilege(
        self,
        role_name: str,
        object: str,
        object_name: str,
        privilege: str,
        db_name: str,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_request(
            role_name,
            object,
            object_name,
            privilege,
            db_name,
            milvus_types.OperatePrivilegeType.Revoke,
        )
        resp = await self._async_stub.OperatePrivilege(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def grant_privilege_v2(
        self,
        role_name: str,
        privilege: str,
        collection_name: str,
        db_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_v2_request(
            role_name,
            privilege,
            milvus_types.OperatePrivilegeType.Grant,
            db_name,
            collection_name,
        )
        resp = await self._async_stub.OperatePrivilegeV2(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def revoke_privilege_v2(
        self,
        role_name: str,
        privilege: str,
        collection_name: str,
        db_name: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        await self.ensure_channel_ready()
        req = Prepare.operate_privilege_v2_request(
            role_name,
            privilege,
            milvus_types.OperatePrivilegeType.Revoke,
            db_name,
            collection_name,
        )
        resp = await self._async_stub.OperatePrivilegeV2(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def create_resource_group(self, name: str, timeout: Optional[float] = None, **kwargs):
        req = Prepare.create_resource_group(name, **kwargs)
        resp = await self._async_stub.CreateResourceGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def drop_resource_group(self, name: str, timeout: Optional[float] = None, **kwargs):
        req = Prepare.drop_resource_group(name)
        resp = await self._async_stub.DropResourceGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def update_resource_groups(
        self, configs: Dict[str, ResourceGroupConfig], timeout: Optional[float] = None, **kwargs
    ):
        req = Prepare.update_resource_groups(configs)
        resp = await self._async_stub.UpdateResourceGroups(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def describe_resource_group(self, name: str, timeout: Optional[float] = None, **kwargs):
        req = Prepare.describe_resource_group(name)
        resp = await self._async_stub.DescribeResourceGroup(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return resp.resource_group

    @retry_on_rpc_failure()
    async def list_resource_groups(self, timeout: Optional[float] = None, **kwargs):
        req = Prepare.list_resource_groups()
        resp = await self._async_stub.ListResourceGroups(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)
        return list(resp.resource_groups)

    @retry_on_rpc_failure()
    async def transfer_replica(
        self,
        source: str,
        target: str,
        collection_name: str,
        num_replica: int,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        req = Prepare.transfer_replica(source, target, collection_name, num_replica)
        resp = await self._async_stub.TransferReplica(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp)

    @retry_on_rpc_failure()
    async def flush(self, collection_names: List[str], timeout: Optional[float] = None, **kwargs):
        req = Prepare.flush_param(collection_names)
        response = await self._async_stub.Flush(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)
        return response

    @retry_on_rpc_failure()
    async def compact(
        self,
        collection_name: str,
        is_clustering: Optional[bool] = False,
        is_l0: Optional[bool] = False,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> int:
        meta = _api_level_md(**kwargs)
        request = Prepare.describe_collection_request(collection_name)
        response = await self._async_stub.DescribeCollection(
            request, timeout=timeout, metadata=meta
        )
        check_status(response.status)

        req = Prepare.manual_compaction(
            collection_name, is_clustering, is_l0, response.collectionID
        )
        response = await self._async_stub.ManualCompaction(req, timeout=timeout, metadata=meta)
        check_status(response.status)

        return response.compactionID

    @retry_on_rpc_failure()
    async def get_compaction_state(
        self, compaction_id: int, timeout: Optional[float] = None, **kwargs
    ):
        req = Prepare.get_compaction_state(compaction_id)
        response = await self._async_stub.GetCompactionState(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(response.status)

        return CompactionState(
            compaction_id,
            State.new(response.state),
            response.executingPlanNo,
            response.timeoutPlanNo,
            response.completedPlanNo,
        )

    @retry_on_rpc_failure()
    async def run_analyzer(
        self,
        texts: Union[str, List[str]],
        analyzer_params: Optional[Union[str, Dict]] = None,
        with_hash: bool = False,
        with_detail: bool = False,
        collection_name: Optional[str] = None,
        field_name: Optional[str] = None,
        analyzer_names: Optional[Union[str, List[str]]] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ):
        req = Prepare.run_analyzer(
            texts,
            analyzer_params=analyzer_params,
            with_hash=with_hash,
            with_detail=with_detail,
            collection_name=collection_name,
            field_name=field_name,
            analyzer_names=analyzer_names,
        )
        resp = await self._async_stub.RunAnalyzer(
            req, timeout=timeout, metadata=_api_level_md(**kwargs)
        )
        check_status(resp.status)

        if isinstance(texts, str):
            return AnalyzeResult(resp.results[0], with_hash, with_detail)
        return [AnalyzeResult(result, with_hash, with_detail) for result in resp.results]



import numpy as np
import time 
import random

from pymilvus import (
    connections,
    list_collections,
    FieldSchema, CollectionSchema, DataType,
    Collection,
)

dim = 128
nb = 10000
nq = 10
collection_name = "GPU_CAGRA_test"

# configure milvus hostname and port
print(f"\nCreate connection...")
connections.connect(host="localhost", port=19530)

# List all collection names
print(f"\nList collections...")
collection_list = list_collections()
print(list_collections())

if(collection_list.count(collection_name)):
    print(collection_name, " exist, and drop it")
    collection = Collection(collection_name)
    collection.drop()
    print("drop")

field1 = FieldSchema(name="id", dtype=DataType.INT64, description="int64", is_primary=True)
field2 = FieldSchema(name = "vec", dtype = DataType.FLOAT_VECTOR, description = "float vector", dim = dim, is_primary = False)
schema = CollectionSchema(fields = [field1, field2], description = "sift decription")
collection = Collection(name = collection_name, data = None, schema = schema, shards_num = 2)

print(list_collections())

print(f"\nList partitions...")
print(collection.partitions)

print("begin insert...")
rng = np.random.default_rng(seed=19530)
data = rng.random((nb, dim))
counter = 0
block_num = 100
block_size = int(data.shape[0]/block_num)
start = time.time()
for t in range(block_num):
    entities = [
            [i for i in range(counter, counter + block_size)],
            # [vectors[i] for i in range(counter, counter + block_size)]
            [vec for vec in data[counter: counter + block_size]]
            ]
    insert_result =  collection.insert(entities)
    counter = counter + block_size
print ("end of insert, cost: ", time.time()-start)

collection.flush()
print(collection.num_entities)

# create index
print(f"\nCreate index...")

collection.create_index(field_name="vec",
        index_params={'index_type': 'GPU_CAGRA',  
            'metric_type': 'L2',
            'params': {
                'intermediate_graph_degree':64,
                'graph_degree': 32,
                }})
print(f"\nCreated index done.")

# load
print(f"\nLoad...")
collection.load()
print(f"\nLoaded")

print(f"\nSearch...")
res = collection.search([ vec for vec in data[0:1]],
                        "vec", 
                            {"metric_type": "L2",
                              "params": {
                                   "search_width":100},
                                   }, 
                                limit=100)
print("run result: ", res[0].ids)
collection.drop()



from pymilvus import MilvusClient, DataType, Function, FunctionType

# 1. Setup Milvus Client
client = MilvusClient("http://localhost:19530")
COLLECTION_NAME = "multilingual_test_A"
if client.has_collection(collection_name=COLLECTION_NAME):
    client.drop_collection(collection_name=COLLECTION_NAME)

# 2. Define analyzers for multiple languages
# These individual analyzer definitions will be reused by both methods.
analyzers = {
    "Japanese": { 
        # Use lindera with japanese dict 'ipadic' 
        # and remove punctuation beacuse lindera tokenizer will remain punctuation
        "tokenizer":{
            "type": "lindera",
            "dict_kind": "ipadic"
        },
        "filter": ["removepunct"]
    },
    "English": {
        # Use build-in english analyzer
        "type": "english",
    },
    "default": {
        # use icu tokenizer as a fallback.
        "tokenizer": "icu",
    }
}

# --- Option A: Using Multi-Language Analyzer ---
print("\n--- Demonstrating Multi-Language Analyzer ---")

# 3A. reate a collection with the Multi Analyzer

mutil_analyzer_params = {
    "by_field": "language",
    "analyzers": analyzers,
}

schema = MilvusClient.create_schema(
    auto_id=True,
    enable_dynamic_field=False,
)
schema.add_field(field_name="id", datatype=DataType.INT64, is_primary=True)# Apply our multi-language analyzer to the 'title' field
schema.add_field(field_name="language", datatype=DataType.VARCHAR, max_length=255, nullable = True)
schema.add_field(field_name="text", datatype=DataType.VARCHAR, max_length=25565, enable_analyzer=True, multi_analyzer_params = mutil_analyzer_params)
schema.add_field(field_name="text_sparse", datatype=DataType.SPARSE_FLOAT_VECTOR) # Bm25 Sparse Vector

# add bm25 function
text_bm25_function = Function(
    name="text_bm25",
    function_type=FunctionType.BM25,
    input_field_names=["text"],
    output_field_names=["text_sparse"],
)
schema.add_function(text_bm25_function)

index_params = client.prepare_index_params()
index_params.add_index(
    field_name="text_sparse",
    index_type="AUTOINDEX", # Use auto index for BM25
    metric_type="BM25",
)

client.create_collection(
    collection_name=COLLECTION_NAME,
    schema=schema,
    index_params=index_params
)
print(f"Collection '{COLLECTION_NAME}' created successfully.")

# 4A. Insert data for Multi-Language Analyzer and load collection# Insert English and Japanese movie titles, explicitly setting the 'language' field
client.insert(
    collection_name=COLLECTION_NAME,
    data=[
        {"text": "The Lord of the Rings", "language": "English"},
        {"text": "Spirited Away", "language": "English"},
        {"text": "千と千尋の神隠し", "language": "Japanese"}, # This is "Spirited Away" in Japanese
        {"text": "君の名は。", "language": "Japanese"}, # This is "Your Name." in Japanese
    ]
)
print(f"Inserted multilingual data into '{COLLECTION_NAME}'.")

# Load the collection into memory before searching
client.load_collection(collection_name=COLLECTION_NAME)

# 5A. Perform a full-text search with Multi-Language Analyzer# When searching, explicitly specify the analyzer to use for the query string.
print("\n--- Search results for Multi-Language Analyzer ---")
results_multi_jp = client.search(
    collection_name=COLLECTION_NAME,
    data=["神隠し"],
    limit=2,
    output_fields=["text"],
    search_params={"metric_type": "BM25", "analyzer_name": "Japanese"}, # Specify Japanese analyzer for query
    consistency_level = "Strong",
)
print("\nSearch results for '神隠し' (Multi-Language Analyzer):")
for result in results_multi_jp[0]:
    print(result)

results_multi_en = client.search(
    collection_name=COLLECTION_NAME,
    data=["Rings"],
    limit=2,
    output_fields=["text"],
    search_params={"metric_type": "BM25", "analyzer_name": "English"}, # Specify English analyzer for query
    consistency_level = "Strong",
)
print("\nSearch results for 'Rings' (Multi-Language Analyzer):")
for result in results_multi_en[0]:
    print(result)

client.drop_collection(collection_name=COLLECTION_NAME)
print(f"Collection '{COLLECTION_NAME}' dropped.")


from pymilvus import (
    DataType,
    MilvusClient,
    AsyncMilvusClient,
    AnnSearchRequest,
    RRFRanker,
)
import numpy as np
import asyncio
import time
import random

fmt = "\n=== {:30} ===\n"
num_entities, dim = 100, 8
default_limit = 3
collection_name = "hello_milvus"
rng = np.random.default_rng(seed=19530)

milvus_client = MilvusClient("example.db")
async_milvus_client = AsyncMilvusClient("example.db")

loop = asyncio.get_event_loop()

schema = milvus_client.create_schema(auto_id=False, description="hello_milvus is the simplest demo to introduce the APIs")
schema.add_field("pk", DataType.VARCHAR, is_primary=True, max_length=100)
schema.add_field("random", DataType.DOUBLE)
schema.add_field("embeddings", DataType.FLOAT_VECTOR, dim=dim)
schema.add_field("embeddings2", DataType.FLOAT_VECTOR, dim=dim)

index_params = milvus_client.prepare_index_params()
index_params.add_index(field_name = "embeddings", index_type = "HNSW", metric_type="L2", nlist=128)
index_params.add_index(field_name = "embeddings2",index_type = "HNSW", metric_type="L2", nlist=128)

# Always use `await` when you want to guarantee the execution order of tasks.
async def recreate_collection():
    print(fmt.format("Start dropping collection"))
    await async_milvus_client.drop_collection(collection_name)
    print(fmt.format("Dropping collection done"))
    print(fmt.format("Start creating collection"))
    await async_milvus_client.create_collection(collection_name, schema=schema, index_params=index_params, consistency_level="Strong")
    print(fmt.format("Creating collection done"))

has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    loop.run_until_complete(recreate_collection())
else:
    print(fmt.format("Start creating collection"))
    loop.run_until_complete(async_milvus_client.create_collection(collection_name, schema=schema, index_params=index_params, consistency_level="Strong"))
    print(fmt.format("Creating collection done"))

print(fmt.format("    all collections    "))
print(milvus_client.list_collections())

print(fmt.format(f"schema of collection {collection_name}"))
print(milvus_client.describe_collection(collection_name))

async def async_insert(collection_name):
    entities = [
        # provide the pk field because `auto_id` is set to False
        [str(i) for i in range(num_entities)],
        rng.random(num_entities).tolist(),  # field random, only supports list
        rng.random((num_entities, dim)),  # field embeddings, supports numpy.ndarray and list
        rng.random((num_entities, dim)),  # field embeddings2, supports numpy.ndarray and list
    ]
    rows = [ {"pk": entities[0][i], "random": entities[1][i], "embeddings": entities[2][i], "embeddings2": entities[3][i]} for i in range (num_entities)]
    print(fmt.format("Start async inserting entities"))

    start_time = time.time()
    tasks = []
    for row in rows:
        task = async_milvus_client.insert(collection_name, [row])
        tasks.append(task)
    await asyncio.gather(*tasks)
    end_time = time.time()
    print(fmt.format("Total time: {:.2f} seconds".format(end_time - start_time)))
    print(fmt.format("Async inserting entities done"))

loop.run_until_complete(async_insert(collection_name))

async def other_async_task(collection_name):
    tasks = []
    # search
    random_vector = rng.random((1, dim))
    random_vector2 = rng.random((1, dim))
    task = async_milvus_client.search(collection_name, random_vector, limit=default_limit, output_fields=["pk"], anns_field="embeddings")
    tasks.append(task)
    # hybrid search
    search_param = {
        "data": random_vector,
        "anns_field": "embeddings",
        "param": {"metric_type": "L2"},
        "limit": default_limit,
        "expr": "random > 0.5"}
    req = AnnSearchRequest(**search_param)
    task = async_milvus_client.hybrid_search(collection_name, [req], RRFRanker(), default_limit, output_fields=["pk"])
    tasks.append(task)
    # get
    random_pk = random.randint(0, num_entities - 1)
    task = async_milvus_client.get(collection_name=collection_name, ids=[random_pk])
    tasks.append(task)
    # query
    task = async_milvus_client.query(collection_name=collection_name, filter="", limit=default_limit)
    tasks.append(task)
    # delete
    task = async_milvus_client.delete(collection_name=collection_name, ids=[random_pk])
    tasks.append(task)
    # insert
    task = async_milvus_client.insert(
        collection_name=collection_name,
        data=[{"pk": str(random_pk), "random": random_vector[0][0], "embeddings": random_vector[0], "embeddings2": random_vector[0]}],
    )
    tasks.append(task)
    # upsert
    task = async_milvus_client.upsert(
        collection_name=collection_name,
        data=[{"pk": str(random_pk), "random": random_vector2[0][0], "embeddings": random_vector2[0], "embeddings2": random_vector2[0]}],
    )
    tasks.append(task)

    results = await asyncio.gather(*tasks)
    return results

results = loop.run_until_complete(other_async_task(collection_name))
for r in results:
    print(r)



import abc
import logging
from typing import Any, Dict, List, Optional, Union

import orjson

from pymilvus.exceptions import DataTypeNotMatchException, ExceptionsMessage
from pymilvus.settings import Config

from . import utils
from .constants import DEFAULT_CONSISTENCY_LEVEL, RANKER_TYPE_RRF, RANKER_TYPE_WEIGHTED

# ruff: noqa: F401
# TODO: This is a patch for older version
from .search_result import Hit, Hits, SearchResult
from .types import DataType, FunctionType

logger = logging.getLogger(__name__)


class FieldSchema:
    def __init__(self, raw: Any):
        self._raw = raw

        self.field_id = 0
        self.name = None
        self.is_primary = False
        self.description = None
        self.auto_id = False
        self.type = DataType.UNKNOWN
        self.indexes = []
        self.params = {}
        self.is_partition_key = False
        self.is_dynamic = False
        self.nullable = False
        self.default_value = None
        self.is_function_output = False
        # For array field
        self.element_type = None
        self.is_clustering_key = False
        self.__pack(self._raw)

    def __pack(self, raw: Any):
        self.field_id = raw.fieldID
        self.name = raw.name
        self.is_primary = raw.is_primary_key
        self.description = raw.description
        self.auto_id = raw.autoID
        self.type = DataType(raw.data_type)
        self.is_partition_key = raw.is_partition_key
        self.element_type = DataType(raw.element_type)
        self.is_clustering_key = raw.is_clustering_key
        self.default_value = raw.default_value
        if raw.default_value is not None and raw.default_value.WhichOneof("data") is None:
            self.default_value = None
        self.is_dynamic = raw.is_dynamic
        self.nullable = raw.nullable
        self.is_function_output = raw.is_function_output

        for type_param in raw.type_params:
            if type_param.key == "params":
                try:
                    self.params[type_param.key] = orjson.loads(type_param.value)
                except Exception as e:
                    logger.error(
                        f"FieldSchema::__pack::65::Failed to load JSON type_param.value: {e}, original data: {type_param.value}"
                    )
                    raise
            else:
                if type_param.key in ["mmap.enabled"]:
                    self.params["mmap_enabled"] = (
                        bool(type_param.value) if type_param.value.lower() != "false" else False
                    )
                    continue
                self.params[type_param.key] = type_param.value
                if type_param.key in ["dim"]:
                    self.params[type_param.key] = int(type_param.value)
                if type_param.key in [Config.MaxVarCharLengthKey] and raw.data_type in (
                    DataType.VARCHAR,
                    DataType.ARRAY,
                ):
                    self.params[type_param.key] = int(type_param.value)

                # TO-DO: use constants defined in orm
                if type_param.key in ["max_capacity"] and raw.data_type in (
                    DataType.ARRAY,
                    DataType._ARRAY_OF_VECTOR,
                ):
                    self.params[type_param.key] = int(type_param.value)

        index_dict = {}
        for index_param in raw.index_params:
            if index_param.key == "params":
                try:
                    index_dict[index_param.key] = orjson.loads(index_param.value)
                except Exception as e:
                    logger.error(
                        f"FieldSchema::__pack::92::Failed to load JSON index_param.value: {e}, original data: {index_param.value}"
                    )
                    raise
            else:
                index_dict[index_param.key] = index_param.value

        self.indexes.extend([index_dict])

    def dict(self):
        _dict = {
            "field_id": self.field_id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "params": self.params or {},
        }
        if self.default_value is not None:
            # default_value is nil match this situation
            if self.default_value.WhichOneof("data") is None:
                self.default_value = None
            else:
                _dict["default_value"] = self.default_value

        if self.element_type:
            _dict["element_type"] = self.element_type

        if self.is_partition_key:
            _dict["is_partition_key"] = True
        if self.is_dynamic:
            _dict["is_dynamic"] = True
        if self.auto_id:
            _dict["auto_id"] = True
        if self.nullable:
            _dict["nullable"] = True
        if self.is_primary:
            _dict["is_primary"] = self.is_primary
        if self.is_clustering_key:
            _dict["is_clustering_key"] = True
        if self.is_function_output:
            _dict["is_function_output"] = True
        return _dict


class StructArrayFieldSchema:
    def __init__(self, raw: Any):
        self._raw = raw

        self.name = None
        self.fields = []
        self.description = None

        self.__pack(self._raw)

    def __pack(self, raw: Any):
        self.name = raw.name
        self.field_id = raw.fieldID
        self.description = raw.description
        self.fields = [FieldSchema(f) for f in raw.fields]

    def dict(self):
        return {
            "field_id": self.field_id,
            "name": self.name,
            "description": self.description,
            "type": DataType._ARRAY_OF_STRUCT,
            "fields": [f.dict() for f in self.fields],
        }


class FunctionSchema:
    def __init__(self, raw: Any):
        self._raw = raw

        self.name = None
        self.description = None
        self.type = None
        self.params = {}
        self.input_field_names = []
        self.input_field_ids = []
        self.output_field_names = []
        self.output_field_ids = []
        self.id = 0

        self.__pack(self._raw)

    def __pack(self, raw: Any):
        self.name = raw.name
        self.description = raw.description
        self.id = raw.id
        self.type = FunctionType(raw.type)
        self.params = {}
        for param in raw.params:
            self.params[param.key] = param.value
        self.input_field_names = raw.input_field_names
        self.input_field_ids = raw.input_field_ids
        self.output_field_names = raw.output_field_names
        self.output_field_ids = raw.output_field_ids

    def dict(self):
        return {
            "name": self.name,
            "id": self.id,
            "description": self.description,
            "type": self.type,
            "params": self.params,
            "input_field_names": self.input_field_names,
            "input_field_ids": self.input_field_ids,
            "output_field_names": self.output_field_names,
            "output_field_ids": self.output_field_ids,
        }


class CollectionSchema:
    def __init__(self, raw: Any):
        self._raw = raw

        self.collection_name = None
        self.description = None
        self.params = {}
        self.fields = []
        self.struct_array_fields = []
        self.functions = []
        self.statistics = {}
        self.auto_id = False  # auto_id is not in collection level any more later
        self.aliases = []
        self.collection_id = 0
        self.consistency_level = DEFAULT_CONSISTENCY_LEVEL  # by default
        self.properties = {}
        self.num_shards = 0
        self.num_partitions = 0
        self.enable_dynamic_field = False
        self.created_timestamp = 0
        self.update_timestamp = 0
        if self._raw:
            self.__pack(self._raw)

    def __pack(self, raw: Any):
        self.collection_name = raw.schema.name
        self.description = raw.schema.description
        self.aliases = list(raw.aliases)
        self.collection_id = raw.collectionID
        self.num_shards = raw.shards_num
        self.num_partitions = raw.num_partitions
        self.created_timestamp = raw.created_timestamp
        self.update_timestamp = raw.update_timestamp
        # keep compatible with older Milvus
        try:
            self.consistency_level = raw.consistency_level
        except Exception:
            self.consistency_level = DEFAULT_CONSISTENCY_LEVEL

        try:
            self.enable_dynamic_field = raw.schema.enable_dynamic_field
        except Exception:
            self.enable_dynamic_field = False

        # TODO: extra_params here
        # for kv in raw.extra_params:

        self.fields = [FieldSchema(f) for f in raw.schema.fields]
        self.struct_array_fields = [
            StructArrayFieldSchema(f) for f in raw.schema.struct_array_fields
        ]
        self.functions = [FunctionSchema(f) for f in raw.schema.functions]
        function_output_field_names = [f for fn in self.functions for f in fn.output_field_names]
        for field in self.fields:
            if field.name in function_output_field_names:
                field.is_function_output = True
        # for s in raw.statistics:

        for p in raw.properties:
            self.properties[p.key] = p.value

    @classmethod
    def _rewrite_schema_dict(cls, schema_dict: Dict):
        fields = schema_dict.get("fields", [])
        if not fields:
            return

        for field_dict in fields:
            if field_dict.get("auto_id", None) is not None:
                schema_dict["auto_id"] = field_dict["auto_id"]
                return

    def dict(self):
        if not self._raw:
            return {}
        _dict = {
            "collection_name": self.collection_name,
            "auto_id": self.auto_id,
            "num_shards": self.num_shards,
            "description": self.description,
            "fields": [f.dict() for f in self.fields],
            "struct_array_fields": [f.dict() for f in self.struct_array_fields],
            "functions": [f.dict() for f in self.functions],
            "aliases": self.aliases,
            "collection_id": self.collection_id,
            "consistency_level": self.consistency_level,
            "properties": self.properties,
            "num_partitions": self.num_partitions,
            "enable_dynamic_field": self.enable_dynamic_field,
        }

        if self.created_timestamp != 0:
            _dict["created_timestamp"] = self.created_timestamp
        if self.update_timestamp != 0:
            _dict["update_timestamp"] = self.update_timestamp
        self._rewrite_schema_dict(_dict)
        return _dict

    def __str__(self):
        return self.dict().__str__()


class MutationResult:
    def __init__(self, raw: Any):
        self._raw = raw
        self._primary_keys = []
        self._insert_cnt = 0
        self._delete_cnt = 0
        self._upsert_cnt = 0
        self._timestamp = 0
        self._succ_index = []
        self._err_index = []
        self._cost = 0

        self._pack(raw)

    @property
    def primary_keys(self):
        return self._primary_keys

    @property
    def insert_count(self):
        return self._insert_cnt

    @property
    def delete_count(self):
        return self._delete_cnt

    @property
    def upsert_count(self):
        return self._upsert_cnt

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def succ_count(self):
        return len(self._succ_index)

    @property
    def err_count(self):
        return len(self._err_index)

    @property
    def succ_index(self):
        return self._succ_index

    @property
    def err_index(self):
        return self._err_index

    # The unit of this cost is vcu, similar to token
    @property
    def cost(self):
        return self._cost

    def __str__(self):
        if self.cost:
            return (
                f"(insert count: {self._insert_cnt}, delete count: {self._delete_cnt}, upsert count: {self._upsert_cnt}, "
                f"timestamp: {self._timestamp}, success count: {self.succ_count}, err count: {self.err_count}, "
                f"cost: {self._cost})"
            )
        return (
            f"(insert count: {self._insert_cnt}, delete count: {self._delete_cnt}, upsert count: {self._upsert_cnt}, "
            f"timestamp: {self._timestamp}, success count: {self.succ_count}, err count: {self.err_count}"
        )

    __repr__ = __str__

    # TODO
    # def error_code(self):
    #     pass
    #
    # def error_reason(self):
    #     pass

    def _pack(self, raw: Any):
        which = raw.IDs.WhichOneof("id_field")
        if which == "int_id":
            self._primary_keys = raw.IDs.int_id.data
        elif which == "str_id":
            self._primary_keys = raw.IDs.str_id.data

        self._insert_cnt = raw.insert_cnt
        self._delete_cnt = raw.delete_cnt
        self._upsert_cnt = raw.upsert_cnt
        self._timestamp = raw.timestamp
        self._succ_index = raw.succ_index
        self._err_index = raw.err_index
        self._cost = int(
            raw.status.extra_info["report_value"] if raw.status and raw.status.extra_info else "0"
        )


class BaseRanker:
    def __int__(self):
        return

    def dict(self):
        return {}

    def __str__(self):
        return self.dict().__str__()


class RRFRanker(BaseRanker):
    def __init__(
        self,
        k: int = 60,
    ):
        self._strategy = RANKER_TYPE_RRF
        self._k = k

    def dict(self):
        params = {
            "k": self._k,
        }
        return {
            "strategy": self._strategy,
            "params": params,
        }


class WeightedRanker(BaseRanker):
    def __init__(self, *nums, norm_score: bool = True):
        self._strategy = RANKER_TYPE_WEIGHTED
        weights = []
        for num in nums:
            # isinstance(True, int) is True, thus we need to check bool first
            if isinstance(num, bool) or not isinstance(num, (int, float)):
                error_msg = f"Weight must be a number, got {type(num)}"
                raise TypeError(error_msg)
            weights.append(num)
        self._weights = weights
        self._norm_score = norm_score

    def dict(self):
        params = {
            "weights": self._weights,
            "norm_score": self._norm_score,
        }
        return {
            "strategy": self._strategy,
            "params": params,
        }


class AnnSearchRequest:
    def __init__(
        self,
        data: Union[List, utils.SparseMatrixInputType],
        anns_field: str,
        param: Dict,
        limit: int,
        expr: Optional[str] = None,
        expr_params: Optional[dict] = None,
    ):
        self._data = data
        self._anns_field = anns_field
        self._param = param
        self._limit = limit

        if expr is not None and not isinstance(expr, str):
            raise DataTypeNotMatchException(message=ExceptionsMessage.ExprType % type(expr))
        self._expr = expr
        self._expr_params = expr_params

    @property
    def data(self):
        return self._data

    @property
    def anns_field(self):
        return self._anns_field

    @property
    def param(self):
        return self._param

    @property
    def limit(self):
        return self._limit

    @property
    def expr(self):
        return self._expr

    @property
    def expr_params(self):
        return self._expr_params

    def __str__(self):
        return {
            "anns_field": self.anns_field,
            "param": self.param,
            "limit": self.limit,
            "expr": self.expr,
        }.__str__()


class LoopBase:
    def __init__(self):
        self.__index = 0

    def __iter__(self):
        return self

    def __getitem__(self, item: Any):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            return [self.get__item(i) for i in range(_start, _end, _step)]

        if item >= self.__len__():
            msg = "Index out of range"
            raise IndexError(msg)

        return self.get__item(item)

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        # iterate stop, raise Exception
        self.__index = 0
        raise StopIteration

    def __str__(self):
        return str(list(map(str, self.__getitem__(slice(0, 10)))))

    @abc.abstractmethod
    def get__item(self, item: Any):
        raise NotImplementedError



import logging
import re
import subprocess
from contextlib import suppress
from importlib.metadata import PackageNotFoundError, version

log = logging.getLogger(__name__)


__version__ = "0.0.0.dev"


with suppress(PackageNotFoundError):
    __version__ = version("pymilvus")


def get_commit(version: str = "", short: bool = True) -> str:
    """get commit return the commit for a specific version like `xxxxxx.dev12`"""

    version_info = r"((\d+)\.(\d+)\.(\d+))((rc)(\d+))?(\.dev(\d+))?"
    # 2.0.0rc9.dev12
    p = re.compile(version_info)

    target_v = __version__ if version == "" else version
    match = p.match(target_v)

    if match is not None:
        match_version = match.groups()
        if match_version[7] is not None:
            if match_version[4] is not None:
                v = str(int(match_version[6]) - 1)
                target_tag = "v" + match_version[0] + match_version[5] + v
            else:
                target_tag = "v" + ".".join(str(int("".join(match_version[1:4])) - 1).split(""))
            target_num = int(match_version[-1])
        elif match_version[4] is not None:
            target_tag = "v" + match_version[0] + match_version[4]
            target_num = 0
        else:
            target_tag = "v" + match_version[0]
            target_num = 0
    else:
        return f"Version: {target_v} isn't the right form"

    try:
        cmd = ["git", "rev-list", "--reverse", "--ancestry-path", f"{target_tag}^..HEAD"]
        log.info(f"git cmd: {' '.join(cmd)}")
        result = subprocess.check_output(cmd).decode("ascii").strip().split("\n")

        length = 7 if short else 40
        return result[target_num][:length]
    except Exception as e:
        return f"Get commit for version {target_v} wrong: {e}"



import time
import numpy as np
from pymilvus import (
    MilvusClient,
    DataType
)

fmt = "\n=== {:30} ===\n"
dim = 8
collection_name = "hello_milvus"
milvus_client = MilvusClient("http://localhost:19530")

has_collection = milvus_client.has_collection(collection_name, timeout=5)
if has_collection:
    milvus_client.drop_collection(collection_name)

schema = milvus_client.create_schema(enable_dynamic_field=True)
schema.add_field("id", DataType.INT64, is_primary=True)
schema.add_field("embeddings", DataType.FLOAT_VECTOR, dim=dim)
schema.add_field("title", DataType.VARCHAR, max_length=64)

# collection is not loaded after creation
milvus_client.create_collection(collection_name, schema=schema, consistency_level="Strong")

rng = np.random.default_rng(seed=19530)
rows = [
        {"id": 1, "embeddings": rng.random((1, dim))[0], "a": 100, "title": "t1"},
        {"id": 2, "embeddings": rng.random((1, dim))[0], "b": 200, "title": "t2"},
        {"id": 3, "embeddings": rng.random((1, dim))[0], "c": 300, "title": "t3"},
        {"id": 4, "embeddings": rng.random((1, dim))[0], "d": 400, "title": "t4"},
        {"id": 5, "embeddings": rng.random((1, dim))[0], "e": 500, "title": "t5"},
        {"id": 6, "embeddings": rng.random((1, dim))[0], "f": 600, "title": "t6"},
]

print(fmt.format("Start inserting entities"))
insert_result = milvus_client.insert(collection_name, rows)
print(fmt.format("Inserting entities done"))
print(insert_result)

index_params = milvus_client.prepare_index_params()
index_params.add_index(field_name = "embeddings", metric_type="L2")
index_params.add_index(field_name = "title", index_type = "Trie", index_name="my_trie")

print(fmt.format("Start create index"))
milvus_client.create_index(collection_name, index_params)


index_names = milvus_client.list_indexes(collection_name)
print(f"index names for {collection_name}:", index_names)
for index_name in index_names:
    index_info = milvus_client.describe_index(collection_name, index_name=index_name)
    print(f"index info for index {index_name} is:", index_info)

print(fmt.format("Start load collection"))
milvus_client.load_collection(collection_name)

print(fmt.format("Start query by specifying primary keys"))
query_results = milvus_client.query(collection_name, ids=[2])
print(query_results[0])

print(fmt.format("Start query by specifying filtering expression"))
query_results = milvus_client.query(collection_name, filter= "f == 600 or title == 't2'")
for ret in query_results: 
    print(ret)

vectors_to_search = rng.random((1, dim))
print(fmt.format(f"Start search with retrieve serveral fields."))
result = milvus_client.search(collection_name, vectors_to_search, limit=3, output_fields=["title"])
for hits in result:
    for hit in hits:
        print(f"hit: {hit}")



field_index_names = milvus_client.list_indexes(collection_name, field_name = "embeddings")
print(f"index names for {collection_name}`s field embeddings:", field_index_names)

try:
    milvus_client.drop_index(collection_name, "my_trie")
except Exception as e:
    print(f"cacthed {e}")

milvus_client.release_collection(collection_name)

milvus_client.drop_index(collection_name, "my_trie")

milvus_client.drop_collection(collection_name)



from typing import Any
from unittest.mock import MagicMock, patch

import grpc
import pytest
from pymilvus import CollectionSchema, FieldSchema, MilvusException
from pymilvus.client.grpc_handler import GrpcHandler
from pymilvus.exceptions import ParamError
from pymilvus.grpc_gen import common_pb2, milvus_pb2
from pymilvus.orm.types import DataType

descriptor = milvus_pb2.DESCRIPTOR.services_by_name["MilvusService"]


class TestGrpcHandler:
    @pytest.mark.parametrize("has", [True, False])
    def test_has_collection_no_error(self, channel, client_thread, has):
        handler = GrpcHandler(channel=channel)

        has_collection_future = client_thread.submit(handler.has_collection, "fake")

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DescribeCollection"]
        )
        rpc.send_initial_metadata(())

        reason = "" if has else "can't find collection"
        code = 0 if has else 100

        expected_result = milvus_pb2.DescribeCollectionResponse(
            status=common_pb2.Status(code=code, reason=reason),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        got_result = has_collection_future.result()
        assert got_result is has

    def test_has_collection_error(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        has_collection_future = client_thread.submit(handler.has_collection, "fake")

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DescribeCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.DescribeCollectionResponse(
            status=common_pb2.Status(code=1, reason="other reason"),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            has_collection_future.result()

    def test_has_collection_Unavailable_exception(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)
        channel.close()

        # Retry is unable to test
        has_collection_future = client_thread.submit(handler.has_collection, "fake", timeout=0)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DescribeCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.DescribeCollectionResponse()

        rpc.terminate(expected_result, (), grpc.StatusCode.UNAVAILABLE, "server Unavailable")

        with pytest.raises(MilvusException):
            has_collection_future.result()

    def test_get_server_version_error(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        get_version_future = client_thread.submit(handler.get_server_version)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["GetVersion"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetVersionResponse(
            status=common_pb2.Status(code=1, reason="unexpected error"),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            get_version_future.result()

    def test_get_server_version(self, channel, client_thread):
        version = "2.2.0"
        handler = GrpcHandler(channel=channel)

        get_version_future = client_thread.submit(handler.get_server_version)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["GetVersion"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetVersionResponse(
            status=common_pb2.Status(code=0),
            version=version,
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        got_result = get_version_future.result()
        assert got_result == version

    @pytest.mark.parametrize("_async", [True])
    def test_flush_all(self, channel, client_thread, _async):
        handler = GrpcHandler(channel=channel)

        flush_all_future = client_thread.submit(handler.flush_all, _async=_async, timeout=10)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["FlushAll"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.FlushAllResponse(
            status=common_pb2.Status(code=0),
            flush_all_ts=100,
        )

        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")
        assert flush_all_future is not None

    def test_get_flush_all_state(self, channel, client_thread):
        handler = GrpcHandler(channel=channel)

        flushed = client_thread.submit(handler.get_flush_all_state, flush_all_ts=100)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["GetFlushAllState"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.GetFlushStateResponse(
            status=common_pb2.Status(code=0),
            flushed=True,
        )

        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")
        assert flushed.result() is True


class TestGrpcHandlerInitialization:
    def test_init_with_uri(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(uri="http://localhost:19530")
            assert handler.server_address == "localhost:19530"

    def test_init_with_host_port(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(host="localhost", port="19530")
            assert handler.server_address == "localhost:19530"

    def test_init_with_secure_connection(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.secure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(uri="http://localhost:19530", secure=True)
            assert handler.server_address == "localhost:19530"

    def test_init_with_invalid_secure_param(self) -> None:
        with pytest.raises(ParamError, match="secure must be bool type"):
            GrpcHandler(uri="http://localhost:19530", secure="not_bool")

    def test_init_with_authorization(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(
                uri="http://localhost:19530",
                user="test_user",
                password="test_password"
            )
            assert handler.server_address == "localhost:19530"

    def test_init_with_token(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(
                uri="http://localhost:19530",
                token="test_token"
            )
            assert handler.server_address == "localhost:19530"

    def test_init_with_db_name(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(
                uri="http://localhost:19530",
                db_name="test_db"
            )
            assert handler.server_address == "localhost:19530"

    def test_get_server_type(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(uri="http://localhost:19530")
            # get_server_type will return 'milvus' for localhost
            server_type = handler.get_server_type()
            assert server_type == "milvus"


class TestGrpcHandlerStateManagement:
    def test_register_state_change_callback(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_ch = MagicMock()
            mock_channel.return_value = mock_ch
            handler = GrpcHandler(uri="http://localhost:19530")

            def callback(state: Any) -> None:
                pass

            handler.register_state_change_callback(callback)
            assert callback in handler.callbacks
            mock_ch.subscribe.assert_called_once_with(callback, try_to_connect=True)

    def test_deregister_state_change_callbacks(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_ch = MagicMock()
            mock_channel.return_value = mock_ch
            handler = GrpcHandler(uri="http://localhost:19530")

            def callback(state: Any) -> None:
                pass

            handler.register_state_change_callback(callback)
            handler.deregister_state_change_callbacks()

            assert len(handler.callbacks) == 0
            mock_ch.unsubscribe.assert_called_once_with(callback)

    def test_close(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_ch = MagicMock()
            mock_channel.return_value = mock_ch
            handler = GrpcHandler(uri="http://localhost:19530")

            # Register a callback first
            def callback(state: Any) -> None:
                pass
            handler.register_state_change_callback(callback)

            handler.close()
            # Verify close was called on the channel
            mock_ch.close.assert_called_once()

    def test_reset_db_name(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(uri="http://localhost:19530")

            # Add some dummy data to schema_cache
            handler.schema_cache["test_collection"] = {"field": "value"}

            with patch.object(handler, '_setup_identifier_interceptor'):
                handler.reset_db_name("new_db")

            assert len(handler.schema_cache) == 0

    def test_set_onetime_loglevel(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_channel.return_value = MagicMock()
            handler = GrpcHandler(uri="http://localhost:19530")

            # Test passes if no exception is raised
            handler.set_onetime_loglevel("DEBUG")

    def test_wait_for_channel_ready_success(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_ch = MagicMock()
            mock_channel.return_value = mock_ch
            handler = GrpcHandler(uri="http://localhost:19530")

            with patch('pymilvus.client.grpc_handler.grpc.channel_ready_future') as mock_future:
                mock_result = MagicMock()
                mock_result.result.return_value = None
                mock_future.return_value = mock_result

                with patch.object(handler, '_setup_identifier_interceptor'):
                    handler._wait_for_channel_ready(timeout=10)

                mock_future.assert_called_once_with(mock_ch)

    def test_wait_for_channel_ready_timeout(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.insecure_channel') as mock_channel:
            mock_ch = MagicMock()
            mock_channel.return_value = mock_ch
            handler = GrpcHandler(uri="http://localhost:19530")

            with patch('pymilvus.client.grpc_handler.grpc.channel_ready_future') as mock_future:
                mock_result = MagicMock()
                mock_result.result.side_effect = grpc.FutureTimeoutError()
                mock_future.return_value = mock_result

                with pytest.raises(MilvusException) as exc_info:
                    handler._wait_for_channel_ready(timeout=10)

                assert "Fail connecting to server" in str(exc_info.value)

    def test_wait_for_channel_ready_no_channel(self) -> None:
        handler = GrpcHandler(channel=None)
        # Manually set channel to None to test the error case
        handler._channel = None

        with pytest.raises(MilvusException) as exc_info:
            handler._wait_for_channel_ready()

        assert "No channel in handler" in str(exc_info.value)


class TestGrpcHandlerCollectionOperations:
    def test_create_collection_sync(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        schema = CollectionSchema([
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128)
        ])

        create_future = client_thread.submit(
            handler.create_collection,
            collection_name="test_collection",
            fields=schema,
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["CreateCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = create_future.result()
        assert result is None

    def test_create_collection_async(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        schema = CollectionSchema([
            FieldSchema(name="id", dtype=DataType.INT64, is_primary=True),
            FieldSchema(name="vector", dtype=DataType.FLOAT_VECTOR, dim=128)
        ])

        create_future = client_thread.submit(
            handler.create_collection,
            collection_name="test_collection",
            fields=schema,
            timeout=10,
            _async=True
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["CreateCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = create_future.result()
        assert result is not None  # Should return a future object

    def test_drop_collection(self, channel: Any, client_thread: Any) -> None:
        """Test drop_collection"""
        handler = GrpcHandler(channel=channel)

        drop_future = client_thread.submit(
            handler.drop_collection,
            collection_name="test_collection",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DropCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = drop_future.result()
        assert result is None

    def test_add_collection_field(self, channel: Any, client_thread: Any) -> None:
        """Test add_collection_field"""
        handler = GrpcHandler(channel=channel)

        field_schema = FieldSchema(name="new_field", dtype=DataType.INT64)

        add_field_future = client_thread.submit(
            handler.add_collection_field,
            collection_name="test_collection",
            field_schema=field_schema,
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["AddCollectionField"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = add_field_future.result()
        assert result is None

    def test_alter_collection_properties(self, channel: Any, client_thread: Any) -> None:
        """Test alter_collection_properties"""
        handler = GrpcHandler(channel=channel)

        properties = {"prop1": "value1", "prop2": "value2"}

        alter_future = client_thread.submit(
            handler.alter_collection_properties,
            collection_name="test_collection",
            properties=properties,
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["AlterCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = alter_future.result()
        assert result is None

    def test_alter_collection_field_properties(self, channel: Any, client_thread: Any) -> None:
        """Test alter_collection_field_properties"""
        handler = GrpcHandler(channel=channel)

        field_params = {"param1": "value1"}

        alter_field_future = client_thread.submit(
            handler.alter_collection_field_properties,
            collection_name="test_collection",
            field_name="test_field",
            field_params=field_params,
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["AlterCollectionField"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = alter_field_future.result()
        assert result is None

    def test_drop_collection_properties(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        property_keys = ["prop1", "prop2"]

        drop_props_future = client_thread.submit(
            handler.drop_collection_properties,
            collection_name="test_collection",
            property_keys=property_keys,
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["AlterCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = drop_props_future.result()
        assert result is None

    def test_has_collection_compatibility(self, channel: Any, client_thread: Any) -> None:
        """Test has_collection with Milvus < 2.3.2 compatibility"""
        handler = GrpcHandler(channel=channel)

        has_collection_future = client_thread.submit(handler.has_collection, "fake")

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DescribeCollection"]
        )
        rpc.send_initial_metadata(())

        # Test compatibility with older Milvus versions
        expected_result = milvus_pb2.DescribeCollectionResponse(
            status=common_pb2.Status(
                error_code=common_pb2.UnexpectedError,
                reason="can't find collection fake"
            ),
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        got_result = has_collection_future.result()
        assert got_result is False


class TestGrpcHandlerPasswordReset:
    def test_reset_password(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        with patch.object(handler, 'update_password') as mock_update:
            with patch.object(handler, '_setup_authorization_interceptor') as mock_auth:
                with patch.object(handler, '_setup_grpc_channel') as mock_setup:
                    handler.reset_password(
                        user="test_user",
                        old_password="old_pass",
                        new_password="new_pass",
                        timeout=10
                    )

                    mock_update.assert_called_once_with(
                        "test_user", "old_pass", "new_pass", timeout=10
                    )
                    mock_auth.assert_called_once_with("test_user", "new_pass", None)
                    mock_setup.assert_called_once()


class TestGrpcHandlerSecureConnection:
    def test_setup_grpc_channel_with_tls(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.secure_channel') as mock_secure:
            with patch('pymilvus.client.grpc_handler.grpc.ssl_channel_credentials') as mock_creds:
                with patch('pymilvus.client.grpc_handler.Path') as mock_path:
                    # Mock file reading
                    mock_file = MagicMock()
                    mock_file.read.return_value = b"cert_content"
                    mock_path.return_value.open.return_value.__enter__.return_value = mock_file

                    mock_secure.return_value = MagicMock()
                    mock_creds.return_value = MagicMock()

                    GrpcHandler(
                        uri="http://localhost:19530",
                        secure=True,
                        server_pem_path="/path/to/server.pem"
                    )

                    # Verify secure channel was created
                    mock_creds.assert_called_once()
                    mock_secure.assert_called_once()

    def test_setup_grpc_channel_with_client_certs(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.secure_channel') as mock_secure:
            with patch('pymilvus.client.grpc_handler.grpc.ssl_channel_credentials') as mock_creds:
                with patch('pymilvus.client.grpc_handler.Path') as mock_path:
                    # Mock file reading
                    mock_file = MagicMock()
                    mock_file.read.return_value = b"cert_content"
                    mock_path.return_value.open.return_value.__enter__.return_value = mock_file

                    mock_secure.return_value = MagicMock()
                    mock_creds.return_value = MagicMock()

                    GrpcHandler(
                        uri="http://localhost:19530",
                        secure=True,
                        client_pem_path="/path/to/client.pem",
                        client_key_path="/path/to/client.key",
                        ca_pem_path="/path/to/ca.pem"
                    )

                    # Verify secure channel was created
                    mock_creds.assert_called_once()
                    mock_secure.assert_called_once()

    def test_setup_grpc_channel_with_server_name_override(self) -> None:
        with patch('pymilvus.client.grpc_handler.grpc.secure_channel') as mock_secure:
            with patch('pymilvus.client.grpc_handler.grpc.ssl_channel_credentials') as mock_creds:
                mock_secure.return_value = MagicMock()
                mock_creds.return_value = MagicMock()

                GrpcHandler(
                    uri="http://localhost:19530",
                    secure=True,
                    server_name="custom.server.name"
                )

                # Check that the server name override was added to options
                call_args = mock_secure.call_args
                options = call_args[1]['options']
                assert any(
                    opt[0] == "grpc.ssl_target_name_override" and opt[1] == "custom.server.name"
                    for opt in options
                )


class TestGrpcHandlerListAndRenameOperations:
    @pytest.mark.parametrize("result", [["collection1", "collection2", "collection3"], []])
    def test_list_collections(self, channel: grpc.Channel, client_thread: Any, result) -> None:
        handler = GrpcHandler(channel=channel)

        list_future = client_thread.submit(handler.list_collections, timeout=10)

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["ShowCollections"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.ShowCollectionsResponse(
            status=common_pb2.Status(code=0),
            collection_names=result
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        assert list_future.result() == result

    def test_list_collections_error(self, channel: grpc.Channel, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)
        list_future = client_thread.submit(handler.list_collections, timeout=10)

        (_, _, rpc) = channel.take_unary_unary(descriptor.methods_by_name["ShowCollections"])
        rpc.send_initial_metadata(())

        rpc.terminate(
            milvus_pb2.ShowCollectionsResponse(status=common_pb2.Status(code=1, reason="Internal error")),
            (),
            grpc.StatusCode.OK,
            "",
        )

        with pytest.raises(MilvusException):
            list_future.result()

    def test_rename_collections(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        rename_future = client_thread.submit(
            handler.rename_collections,
            old_name="old_collection",
            new_name="new_collection",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["RenameCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = rename_future.result()
        assert result is None

    def test_rename_collections_with_new_db(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        rename_future = client_thread.submit(
            handler.rename_collections,
            old_name="old_collection",
            new_name="new_collection",
            new_db_name="new_database",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["RenameCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = rename_future.result()
        assert result is None

    def test_rename_collections_error(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        rename_future = client_thread.submit(
            handler.rename_collections,
            old_name="old_collection",
            new_name="new_collection",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["RenameCollection"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=1, reason="Collection already exists")
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            rename_future.result()


class TestGrpcHandlerPartitionOperations:
    def test_create_partition(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        create_future = client_thread.submit(
            handler.create_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["CreatePartition"]
        )
        rpc.send_initial_metadata(())
        rpc.terminate(common_pb2.Status(code=0), (), grpc.StatusCode.OK, "")

        result = create_future.result()
        assert result is None

    def test_create_partition_error(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        create_future = client_thread.submit(
            handler.create_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["CreatePartition"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=1, reason="Partition already exists")
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            create_future.result()

    def test_drop_partition(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        drop_future = client_thread.submit(
            handler.drop_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DropPartition"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=0)
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = drop_future.result()
        assert result is None

    def test_drop_partition_error(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        drop_future = client_thread.submit(
            handler.drop_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["DropPartition"]
        )
        rpc.send_initial_metadata(())

        expected_result = common_pb2.Status(code=1, reason="Partition not found")
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            drop_future.result()

    @pytest.mark.parametrize("has", [True, False])
    def test_has_partition(self, channel: Any, client_thread: Any, has: bool) -> None:
        handler = GrpcHandler(channel=channel)

        has_future = client_thread.submit(
            handler.has_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["HasPartition"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.BoolResponse(
            status=common_pb2.Status(code=0),
            value=has
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = has_future.result()
        assert result == has

    def test_has_partition_error(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        has_future = client_thread.submit(
            handler.has_partition,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["HasPartition"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.BoolResponse(
            status=common_pb2.Status(code=1, reason="Internal error")
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        with pytest.raises(MilvusException):
            has_future.result()

    def test_list_partitions(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        list_future = client_thread.submit(
            handler.list_partitions,
            collection_name="test_collection",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["ShowPartitions"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.ShowPartitionsResponse(
            status=common_pb2.Status(code=0),
            partition_names=["_default", "partition1", "partition2"]
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = list_future.result()
        assert result == ["_default", "partition1", "partition2"]

    def test_list_partitions_empty(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        list_future = client_thread.submit(
            handler.list_partitions,
            collection_name="test_collection",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["ShowPartitions"]
        )
        rpc.send_initial_metadata(())

        expected_result = milvus_pb2.ShowPartitionsResponse(
            status=common_pb2.Status(code=0),
            partition_names=[]
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = list_future.result()
        assert result == []

    def test_get_partition_stats(self, channel: Any, client_thread: Any) -> None:
        handler = GrpcHandler(channel=channel)

        stats_future = client_thread.submit(
            handler.get_partition_stats,
            collection_name="test_collection",
            partition_name="test_partition",
            timeout=10
        )

        (invocation_metadata, request, rpc) = channel.take_unary_unary(
            descriptor.methods_by_name["GetPartitionStatistics"]
        )
        rpc.send_initial_metadata(())

        stats = [
            common_pb2.KeyValuePair(key="row_count", value="1000")
        ]
        expected_result = milvus_pb2.GetPartitionStatisticsResponse(
            status=common_pb2.Status(code=0),
            stats=stats
        )
        rpc.terminate(expected_result, (), grpc.StatusCode.OK, "")

        result = stats_future.result()
        assert result == stats



# 1. connect to Milvus
# 2. create collection
# 3. insert data
# 4. create index
# 5. query on entities
# 6. drop collection
import time

import numpy as np
from pymilvus import (
    connections,
    utility,
    FieldSchema, CollectionSchema, DataType,
    Collection,
)

fmt = "\n=== {:30} ===\n"
search_latency_fmt = "search latency = {:.4f}s"
dim = 8

#################################################################################
# 1. connect to Milvus
print(fmt.format("start connecting to Milvus"))
connections.connect("default", host="localhost", port="19530")

has = utility.has_collection("hello_milvus")
print(f"Does collection hello_milvus exist in Milvus: {has}")

#################################################################################
# 2. create collection
# We're going to create a collection with 4 fields.
# +-+---------------------+------------+---------------------+--------------------------------------------------+
# | |     field name      | field type |   other attributes  |                  field description               |
# +-+---------------------+------------+---------------------+--------------------------------------------------+
# |1|         "pk"        |   VarChar  |    is_primary=True  |                "primary field"                   |
# | |                     |            |     auto_id=False   |                                                  |
# +-+---------------------+------------+---------------------+--------------------------------------------------+
# |2|    "nullable_fid"   |    Double  |    nullable=True    |        "a double field can insert null"          |
# +-+---------------------+------------+---------------------+--------------------------------------------------+
# |3| "default_value_fid" |    Int64   |   default_value=1   |  "a int64 field can insert with default value"   |
# +-+---------------------+------------+---------------------+--------------------------------------------------+
# |4|     "embeddings"    | FloatVector|        dim=8        |              "float vector with dim 8"           |
# +-+---------------------+------------+---------------------+--------------------------------------------------+
fields = [
    FieldSchema(name="pk", dtype=DataType.VARCHAR, is_primary=True, auto_id=False, max_length=100),
    FieldSchema(name="nullable_fid", dtype=DataType.DOUBLE,nullable=True),
    FieldSchema(name="default_value_fid", dtype=DataType.INT64,default_value=1),
    FieldSchema(name="embeddings", dtype=DataType.FLOAT_VECTOR, dim=dim)
]

schema = CollectionSchema(fields, "hello_milvus is the demo to introduce the nullable and default value functions")

print(fmt.format("Create collection `hello_milvus`"))
hello_milvus = Collection("hello_milvus", schema, consistency_level="Strong")

################################################################################
# 3. insert data
# For fields marked as nullable=True: 
# you can skip the field when inserting data, or set it directly to a null value, and the system will treat it as null 
# For fields marked default_value: 
# the system will automatically apply this value if no data is specified for the field during insertion

print(fmt.format("Start inserting entities"))
rng = np.random.default_rng(seed=19530)

data_rows = [
    {
        # skip the field when inserting data
        "pk": "19530",
        "embeddings": rng.random((1, dim), np.float32)[0]
    },
    {
        # set it directly to a null value
        "pk": "19531",
        "nullable_fid": None,
        "default_value_fid": None,
        "embeddings": rng.random((1, dim), np.float32)[0]
    },
]

hello_milvus.insert(data_rows)

hello_milvus.flush()
print(f"Number of entities in Milvus: {hello_milvus.num_entities}")  # check the num_entities

################################################################################
# 4. create index
print(fmt.format("Start Creating index IVF_FLAT"))
index = {
    "index_type": "IVF_FLAT",
    "metric_type": "L2",
    "params": {"nlist": 128},
}

hello_milvus.create_index("embeddings", index)

################################################################################
# 5. query on entities
print(fmt.format("Start loading"))
hello_milvus.load()

print(fmt.format("Start querying"))
start_time = time.time()
result = hello_milvus.query(expr='pk in ["19530","19531"]', output_fields=["nullable_fid", "default_value_fid","embeddings"])
end_time = time.time()

print(f"query result:\n-{result[0]}")
print(search_latency_fmt.format(end_time - start_time))

###############################################################################
# 6. drop collection
# Finally, drop the hello_milvus collection
print(fmt.format("Drop collection `hello_milvus`"))
utility.drop_collection("hello_milvus")
